use crate::runtime::{Reference, type_key};
use crate::tag;

mod builder; pub use builder::*;
mod util; pub use util::*;

mod varying; pub use varying::Varying;
mod boolean; pub use boolean::Boolean;
mod natural; pub use natural::Natural;
mod integer; pub use integer::Integer;
mod block; pub use block::Block;
mod sequence; pub use sequence::Sequence;
mod array; pub use array::Array;
mod list; pub use list::List;
mod schema; pub use schema::Schema;

pub enum Type {
    Varying(Varying),
    Null,
    Boolean(Boolean),
    Natural(Natural),
    Integer(Integer),
    Block(Block),
    Array(Array),
    List(List),
}
impl Type {
    pub fn from(addr:Reference) -> Self
    {
        match unsafe {type_key(addr.class)} {
            tag::VARYING => Type::Varying(Varying::from(addr).unwrap()),
            tag::NULL => Type::Null,
            tag::BOOLEAN => Type::Boolean(Boolean::from(addr).unwrap()),
            tag::NATURAL => Type::Natural(Natural::from(addr).unwrap()),
            tag::INTEGER => Type::Integer(Integer::from(addr).unwrap()),
            tag::ARRAY => Type::Array(Array::from(addr).unwrap()),
            tag::LIST => Type::List(List::from(addr).unwrap()),
            _ => Type::Null,
        }
    }
}
