use crate::runtime::{
    //util::{name_indexof, name_keyof},
    util::name_indexof,
    Reference,
    acquire, release,
    type_key,
    schema_length,
    schema_add,
    schema_remove,
    schema_reorder,
    schema_map, schema_unmap, schema_index,
    schema_bind,
};
use crate::tag;
use super::schema;

//pub struct Builder {
//    data:Schema,
//}
//impl Builder {
//    pub fn add() { }
//    pub fn map() { }
//    pub fn bind() { }
//}

pub struct Schema {
    managed:bool,
    addr:Reference,
}
impl Schema {
    pub fn new() -> Self
    {
        Self {
            managed:true,
            addr:unsafe {acquire(schema())},
        }
    }

    pub fn from(addr:Reference) -> Result<Self,()>
    {
        return if(unsafe {type_key(addr.class)} == tag::SCHEMA) {
            Ok(Self { managed:false, addr:addr })
        }
        else {
            Err(())
        }
    }

    pub fn with(data:Vec<(&str, usize)>) -> Self
    {
        let mut obj = Self::new();
        for binding in data {
            let (key, type_id) = binding;
            obj.assign(key, type_id);
        }
        return obj;
    }

    //pub fn build() -> Builder
    //{
    //    Builder { }
    //}
    
    pub fn length(&self) -> usize
    {
        unsafe {schema_length(self.addr)}
    }

    pub fn add(&mut self, type_id:usize) -> usize
    {
        unsafe {schema_add(self.addr, type_id)}
    }

    pub fn assign(&mut self, key:&str, type_id:usize)
    {
        if key.len() > 0 {
            let key_index = name_indexof(key);
            unsafe {
                schema_map(
                    self.addr,
                    key_index,
                    schema_add(self.addr, type_id)
                );
            }
        }
        else {
            self.add(type_id);
        }
    }

    pub fn remove(&mut self, index:usize)
    {
        unsafe { schema_remove(self.addr, index); }
    }

    pub fn reorder(&mut self, from:usize, to:usize)
    {
        unsafe { schema_reorder(self.addr, from, to); }
    }

    pub fn map(&mut self, key:&str, index:usize)
    {
        let key_index = name_indexof(key);
        unsafe { schema_map(self.addr, key_index, index); }
    }

    pub fn unmap(&mut self, key:&str)
    {
        let key_index = name_indexof(key);
        unsafe { schema_unmap(self.addr, key_index); }
    }

    pub fn index(&mut self, key:&str) -> usize
    {
        let key_index = name_indexof(key);
        unsafe {schema_index(self.addr, key_index)}
    }

    //pub fn key(&mut self, index:usize) -> String
    //{
    //    name_keyof(unsafe {schema_key(self.addr, index)})
    //}

    pub fn bind(&mut self, id:usize) -> usize
    {
        unsafe {schema_bind(self.addr, id)}
    }
}
impl std::ops::Deref for Schema {
    type Target = Reference;
    fn deref(&self) -> &Self::Target { return &self.addr; }
}
impl Drop for Schema {
    fn drop(&mut self) { if self.managed { unsafe {release(self.addr)}; } }
}
